%for part 2 a kalman filter will be developed for tracking a target moving
%in 2D. The targets motion is defined by kinematic equations with a small
%amount of random motion captured by the acceleration noise as will become
%apparent below, and the sensors are corrupted by additive white gaussian
%noise. the result should demonstrate the filters effectiveness
clear all;
clc;
clf;

N = 500;
Ts = 0.02; %50Hz sampling
Time = 1:1:N;
Time = Time*Ts;

A = [0 1 0 0;
    0 0 0 0;
    0 0 0 1;
    0 0 0 0];

B = [0 1 0 0;
    0 0 0 0.3]'; %use this to input an initial velocity

C = [1 0 0 0;
    0 0 1 0];

D = 0;
sys = ss(A,B,C,D);
sysd = c2d(sys,Ts); %Discretise the continuous state space sys


input = [zeros(5,2);ones(N-5,2)]';  %generating an arbitrary (unmodelled) input - what is the significance??


Q = [0 0 0 0;
    0 0.1 0 0;
    0 0 0 0;
    0 0 0 0.1]; %Covariance state
Q = Q + ones(4)*10^-10;
WQ = Q^(-1/2);
DWQ = inv(WQ);

noiseState = DWQ*randn(4,N); %producing noise with the given covariance matrix

R = [7 0;
    0 7]; %covariance sensors
R = R + ones(2)*10^-10;
WR = R^(-1/2); %this will take a variable with covar mat R to a whitened state
DWR = inv(WR); %dewhiten a white unit power GRV to covar R


noiseSensors = DWR*randn(2,N); %producing noise with the given covariance matrix


Kalman = dsp.KalmanFilter(sysd.A,sysd.C,Q,R,sysd.B,...
            'InitialStateEstimate', [0;0;0;0],...
            'InitialErrorCovarianceEstimate',[0 0 0 0;0 0 0 0;0 0 0 0;0 0 0 0],...
            'ControlInputPort',false); %%this is the kalman filter object. implementing it from scratch is shown below

state = [0;0;0;0]; %True inital state
for i = 2:1:N %%motion simulation
    
    state(:,i) = sysd.A*state(:,i-1) + sysd.B*input(:,i-1) + noiseState(:,i-1);
    Measurement(:,i) = sysd.C*state(:,i-1) + noiseSensors(:,i-1);
    estVal(:,i) = step(Kalman, Measurement(:,i)); %use the Kalman filter to estimate the true state (tracked by the state variable) from the measurements
    
end

figure(1)
hold on
plot(Time,Measurement');
plot(Time,state([1,3],:)','k');

plot(Time,estVal','b--');




title('Matlab Kalman Filter Prediction')
xlabel('Time (s)') % x-axis label
ylabel('Displacement(m)') % y-axis label
legend('Measured Disp in X','Measured Disp in Y','True position X','True position Y','Estimated position X','Estimated position Y');







%%Direct implementation optimal gain KF - directly from wikipedia :D
%this implementation is important to the Extended Kalman filter which
%linearises a non-linear system about the local state at each step - IE the
%locally optimal kalman gain varies with time

estimatedState = [2;-3;1;2]; %%inital state estimate
currentCovariance = [1 0 0 0;0 1 0 0;0 0 1 0;0 0 0 1];

for i = 2:1:N 
    predictedState = sysd.A*estimatedState(:,i-1); %predict positon based on previous state estimate and control effort if known/applicable
    PredictedP = sysd.A*currentCovariance*sysd.A' + Q; %compute the covariance matrix of our prediction - Analgous to the std deviation and how trustworthy our positional estimate is
    
    innovation = Measurement(:,i) - sysd.C*predictedState; %the difference between the measurement, and the predicted, noiseless output
    InnovationCov = sysd.C*PredictedP*(sysd.C)' + R; %the covariance of the innovation, IE how trustworthy our measurement is.
    KalGain = PredictedP*sysd.C'*inv(InnovationCov); %the guts of the algorithm, the kalman gain is basically a weighting between the reliabilty of the prediction, and the reliability of the measurement.
    Kgain(i,:) = [KalGain(:,1);KalGain(:,2)];
    estimatedState(:,i) = predictedState + KalGain*innovation; %Our A-posteriori prediction - which includes the correction to the estimated position based on the measurement
    currentCovariance = (eye(4) - KalGain*sysd.C)*PredictedP; %A-posteriori covariance (garunteed to be no larger than the A-priori covariance)  
end

figure(2)
hold on;
plot(Time,Measurement');
plot(Time,state([1,3],:)','k');
plot(Time,(sysd.C*estimatedState)','b--');

title('Direct implementation with optimal Kalman gain')
xlabel('Time (s)') % x-axis label
ylabel('Displacement(m)') % y-axis label
legend('Measured Disp in X','Measured Disp in Y','True position X','True position Y','Estimated position X','Estimated position Y');




figure(3)
plot(Time,Kgain) %note that the kalman gain converges to a steady state. This is typical. Computing the Kalman Gain is expensive as it has a matrix inverse. Generally its better to find the steady state gain and fix the terms. see below
%For the eqn that computes the steady state gain

title('Kalman gain matrix coefficents')
xlabel('Time (s)') % x-axis label
ylabel('Magnitude') % y-axis label
legend('11','12','21','22');




%%Steady state KF
Lupanyov = zeros(4);
for i =1:1:1000
    Lupanyov = sysd.A*Lupanyov*sysd.A' + Q - sysd.A*Lupanyov*sysd.C'*inv(sysd.C*Lupanyov*sysd.C' + R)*sysd.C*Lupanyov*sysd.A';
end
Gain = sysd.A*Lupanyov*sysd.C'*inv(sysd.C*Lupanyov*sysd.C' + R);


%now the predict->update is much faster:
for i = 2:1:N 
    predictedState = sysd.A*estimatedState(:,i-1); 
    
    innovation = Measurement(:,i) - sysd.C*predictedState;
    estimatedState(:,i) = predictedState + Gain*innovation;  
end

figure(4)
hold on
plot(Time,Measurement');
plot(Time,state([1,3],:)','k');
plot(Time,(sysd.C*estimatedState)','b--');

title('Fixed Kalman Gain')
xlabel('Time (s)') % x-axis label
ylabel('Displacement(m)') % y-axis label
legend('Measured Disp in X','Measured Disp in Y','True position X','True position Y','Estimated position X','Estimated position Y');

